from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.data_classes import (
    ReserveBidTimeSeries,
    ReserveBidTimeSeriesApply,
    ReserveBidTimeSeriesApplyList,
    ReserveBidTimeSeriesFields,
    ReserveBidTimeSeriesList,
    ReserveBidTimeSeriesTextFields,
)
from cognite.powerops.client._generated.data_classes._reserve_bid_time_series import (
    _RESERVEBIDTIMESERIES_PROPERTIES_BY_FIELD,
)

from ._core import DEFAULT_LIMIT_READ, IN_FILTER_LIMIT, Aggregations, TypeAPI


class ReserveBidTimeSeriesPeriodsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str], space="power-ops") -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "ReserveBidTimeSeries.Periods"},
        )
        if isinstance(external_id, str):
            is_reserve_bid_time_series = f.Equals(
                ["edge", "startNode"],
                {"space": space, "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_reserve_bid_time_series)
            )

        else:
            is_reserve_bid_time_series_list = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_reserve_bid_time_series_list)
            )

    def list(
        self, reserve_bid_time_series_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ, space="power-ops"
    ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "ReserveBidTimeSeries.Periods"},
        )
        filters.append(is_edge_type)
        if reserve_bid_time_series_id:
            reserve_bid_time_series_ids = (
                [reserve_bid_time_series_id]
                if isinstance(reserve_bid_time_series_id, str)
                else reserve_bid_time_series_id
            )
            is_reserve_bid_time_series_list = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in reserve_bid_time_series_ids],
            )
            filters.append(is_reserve_bid_time_series_list)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class ReserveBidTimeSeriesAPI(TypeAPI[ReserveBidTimeSeries, ReserveBidTimeSeriesApply, ReserveBidTimeSeriesList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ReserveBidTimeSeries,
            class_apply_type=ReserveBidTimeSeriesApply,
            class_list=ReserveBidTimeSeriesList,
        )
        self._view_id = view_id
        self.periods = ReserveBidTimeSeriesPeriodsAPI(client)

    def apply(
        self,
        reserve_bid_time_series: ReserveBidTimeSeriesApply | Sequence[ReserveBidTimeSeriesApply],
        replace: bool = False,
    ) -> dm.InstancesApplyResult:
        if isinstance(reserve_bid_time_series, ReserveBidTimeSeriesApply):
            instances = reserve_bid_time_series.to_instances_apply()
        else:
            instances = ReserveBidTimeSeriesApplyList(reserve_bid_time_series).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ReserveBidTimeSeries:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ReserveBidTimeSeriesList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> ReserveBidTimeSeries | ReserveBidTimeSeriesList:
        if isinstance(external_id, str):
            reserve_bid_time_series = self._retrieve((self._sources.space, external_id))

            period_edges = self.periods.retrieve(external_id)
            reserve_bid_time_series.periods = [edge.end_node.external_id for edge in period_edges]

            return reserve_bid_time_series
        else:
            reserve_bid_time_series_list = self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

            period_edges = self.periods.retrieve(external_id)
            self._set_periods(reserve_bid_time_series_list, period_edges)

            return reserve_bid_time_series_list

    def search(
        self,
        query: str,
        properties: ReserveBidTimeSeriesTextFields | Sequence[ReserveBidTimeSeriesTextFields] | None = None,
        bid_type: str | list[str] | None = None,
        bid_type_prefix: str | None = None,
        measure_unit: str | list[str] | None = None,
        measure_unit_prefix: str | None = None,
        currency: str | list[str] | None = None,
        currency_prefix: str | None = None,
        min_price: float | None = None,
        max_price: float | None = None,
        direction_name: str | list[str] | None = None,
        direction_name_prefix: str | None = None,
        reserve_object: str | list[str] | None = None,
        reserve_object_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ReserveBidTimeSeriesList:
        filter_ = _create_filter(
            self._view_id,
            bid_type,
            bid_type_prefix,
            measure_unit,
            measure_unit_prefix,
            currency,
            currency_prefix,
            min_price,
            max_price,
            direction_name,
            direction_name_prefix,
            reserve_object,
            reserve_object_prefix,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _RESERVEBIDTIMESERIES_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ReserveBidTimeSeriesFields | Sequence[ReserveBidTimeSeriesFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: ReserveBidTimeSeriesTextFields | Sequence[ReserveBidTimeSeriesTextFields] | None = None,
        bid_type: str | list[str] | None = None,
        bid_type_prefix: str | None = None,
        measure_unit: str | list[str] | None = None,
        measure_unit_prefix: str | None = None,
        currency: str | list[str] | None = None,
        currency_prefix: str | None = None,
        min_price: float | None = None,
        max_price: float | None = None,
        direction_name: str | list[str] | None = None,
        direction_name_prefix: str | None = None,
        reserve_object: str | list[str] | None = None,
        reserve_object_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ReserveBidTimeSeriesFields | Sequence[ReserveBidTimeSeriesFields] | None = None,
        group_by: ReserveBidTimeSeriesFields | Sequence[ReserveBidTimeSeriesFields] = None,
        query: str | None = None,
        search_properties: ReserveBidTimeSeriesTextFields | Sequence[ReserveBidTimeSeriesTextFields] | None = None,
        bid_type: str | list[str] | None = None,
        bid_type_prefix: str | None = None,
        measure_unit: str | list[str] | None = None,
        measure_unit_prefix: str | None = None,
        currency: str | list[str] | None = None,
        currency_prefix: str | None = None,
        min_price: float | None = None,
        max_price: float | None = None,
        direction_name: str | list[str] | None = None,
        direction_name_prefix: str | None = None,
        reserve_object: str | list[str] | None = None,
        reserve_object_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ReserveBidTimeSeriesFields | Sequence[ReserveBidTimeSeriesFields] | None = None,
        group_by: ReserveBidTimeSeriesFields | Sequence[ReserveBidTimeSeriesFields] | None = None,
        query: str | None = None,
        search_property: ReserveBidTimeSeriesTextFields | Sequence[ReserveBidTimeSeriesTextFields] | None = None,
        bid_type: str | list[str] | None = None,
        bid_type_prefix: str | None = None,
        measure_unit: str | list[str] | None = None,
        measure_unit_prefix: str | None = None,
        currency: str | list[str] | None = None,
        currency_prefix: str | None = None,
        min_price: float | None = None,
        max_price: float | None = None,
        direction_name: str | list[str] | None = None,
        direction_name_prefix: str | None = None,
        reserve_object: str | list[str] | None = None,
        reserve_object_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            bid_type,
            bid_type_prefix,
            measure_unit,
            measure_unit_prefix,
            currency,
            currency_prefix,
            min_price,
            max_price,
            direction_name,
            direction_name_prefix,
            reserve_object,
            reserve_object_prefix,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _RESERVEBIDTIMESERIES_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ReserveBidTimeSeriesFields,
        interval: float,
        query: str | None = None,
        search_property: ReserveBidTimeSeriesTextFields | Sequence[ReserveBidTimeSeriesTextFields] | None = None,
        bid_type: str | list[str] | None = None,
        bid_type_prefix: str | None = None,
        measure_unit: str | list[str] | None = None,
        measure_unit_prefix: str | None = None,
        currency: str | list[str] | None = None,
        currency_prefix: str | None = None,
        min_price: float | None = None,
        max_price: float | None = None,
        direction_name: str | list[str] | None = None,
        direction_name_prefix: str | None = None,
        reserve_object: str | list[str] | None = None,
        reserve_object_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            bid_type,
            bid_type_prefix,
            measure_unit,
            measure_unit_prefix,
            currency,
            currency_prefix,
            min_price,
            max_price,
            direction_name,
            direction_name_prefix,
            reserve_object,
            reserve_object_prefix,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _RESERVEBIDTIMESERIES_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        bid_type: str | list[str] | None = None,
        bid_type_prefix: str | None = None,
        measure_unit: str | list[str] | None = None,
        measure_unit_prefix: str | None = None,
        currency: str | list[str] | None = None,
        currency_prefix: str | None = None,
        min_price: float | None = None,
        max_price: float | None = None,
        direction_name: str | list[str] | None = None,
        direction_name_prefix: str | None = None,
        reserve_object: str | list[str] | None = None,
        reserve_object_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> ReserveBidTimeSeriesList:
        filter_ = _create_filter(
            self._view_id,
            bid_type,
            bid_type_prefix,
            measure_unit,
            measure_unit_prefix,
            currency,
            currency_prefix,
            min_price,
            max_price,
            direction_name,
            direction_name_prefix,
            reserve_object,
            reserve_object_prefix,
            external_id_prefix,
            filter,
        )

        reserve_bid_time_series_list = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            if len(external_ids := reserve_bid_time_series_list.as_external_ids()) > IN_FILTER_LIMIT:
                period_edges = self.periods.list(limit=-1)
            else:
                period_edges = self.periods.list(external_ids, limit=-1)
            self._set_periods(reserve_bid_time_series_list, period_edges)

        return reserve_bid_time_series_list

    @staticmethod
    def _set_periods(reserve_bid_time_series_list: Sequence[ReserveBidTimeSeries], period_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in period_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for reserve_bid_time_series in reserve_bid_time_series_list:
            node_id = reserve_bid_time_series.id_tuple()
            if node_id in edges_by_start_node:
                reserve_bid_time_series.periods = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    bid_type: str | list[str] | None = None,
    bid_type_prefix: str | None = None,
    measure_unit: str | list[str] | None = None,
    measure_unit_prefix: str | None = None,
    currency: str | list[str] | None = None,
    currency_prefix: str | None = None,
    min_price: float | None = None,
    max_price: float | None = None,
    direction_name: str | list[str] | None = None,
    direction_name_prefix: str | None = None,
    reserve_object: str | list[str] | None = None,
    reserve_object_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if bid_type and isinstance(bid_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("BidType"), value=bid_type))
    if bid_type and isinstance(bid_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("BidType"), values=bid_type))
    if bid_type_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("BidType"), value=bid_type_prefix))
    if measure_unit and isinstance(measure_unit, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("MeasureUnit"), value=measure_unit))
    if measure_unit and isinstance(measure_unit, list):
        filters.append(dm.filters.In(view_id.as_property_ref("MeasureUnit"), values=measure_unit))
    if measure_unit_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("MeasureUnit"), value=measure_unit_prefix))
    if currency and isinstance(currency, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("Currency"), value=currency))
    if currency and isinstance(currency, list):
        filters.append(dm.filters.In(view_id.as_property_ref("Currency"), values=currency))
    if currency_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("Currency"), value=currency_prefix))
    if min_price or max_price:
        filters.append(dm.filters.Range(view_id.as_property_ref("Price"), gte=min_price, lte=max_price))
    if direction_name and isinstance(direction_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("DirectionName"), value=direction_name))
    if direction_name and isinstance(direction_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("DirectionName"), values=direction_name))
    if direction_name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("DirectionName"), value=direction_name_prefix))
    if reserve_object and isinstance(reserve_object, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("ReserveObject"), value=reserve_object))
    if reserve_object and isinstance(reserve_object, list):
        filters.append(dm.filters.In(view_id.as_property_ref("ReserveObject"), values=reserve_object))
    if reserve_object_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("ReserveObject"), value=reserve_object_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = ["Duration", "DurationApply", "DurationList", "DurationApplyList", "DurationFields", "DurationTextFields"]


DurationTextFields = Literal["unit"]
DurationFields = Literal["duration", "unit"]

_DURATION_PROPERTIES_BY_FIELD = {
    "duration": "duration",
    "unit": "unit",
}


class Duration(DomainModel):
    space: str = "power-ops"
    duration: Optional[int] = None
    unit: Optional[str] = None

    def as_apply(self) -> DurationApply:
        return DurationApply(
            external_id=self.external_id,
            duration=self.duration,
            unit=self.unit,
        )


class DurationApply(DomainModelApply):
    space: str = "power-ops"
    duration: Optional[int] = None
    unit: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.duration is not None:
            properties["duration"] = self.duration
        if self.unit is not None:
            properties["unit"] = self.unit
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "Duration"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class DurationList(TypeList[Duration]):
    _NODE = Duration

    def as_apply(self) -> DurationApplyList:
        return DurationApplyList([node.as_apply() for node in self.data])


class DurationApplyList(TypeApplyList[DurationApply]):
    _NODE = DurationApply



from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.data_classes import (
    OutputContainer,
    OutputContainerApply,
    OutputContainerApplyList,
    OutputContainerFields,
    OutputContainerList,
    OutputContainerTextFields,
)
from cognite.powerops.client._generated.data_classes._output_container import _OUTPUTCONTAINER_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, IN_FILTER_LIMIT, Aggregations, TypeAPI


class OutputContainerMappingsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str], space="power-ops") -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "OutputContainer.mappings"},
        )
        if isinstance(external_id, str):
            is_output_container = f.Equals(
                ["edge", "startNode"],
                {"space": space, "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_output_container)
            )

        else:
            is_output_containers = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_output_containers)
            )

    def list(
        self, output_container_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ, space="power-ops"
    ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "OutputContainer.mappings"},
        )
        filters.append(is_edge_type)
        if output_container_id:
            output_container_ids = (
                [output_container_id] if isinstance(output_container_id, str) else output_container_id
            )
            is_output_containers = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in output_container_ids],
            )
            filters.append(is_output_containers)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class OutputContainerAPI(TypeAPI[OutputContainer, OutputContainerApply, OutputContainerList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=OutputContainer,
            class_apply_type=OutputContainerApply,
            class_list=OutputContainerList,
        )
        self._view_id = view_id
        self.mappings = OutputContainerMappingsAPI(client)

    def apply(
        self, output_container: OutputContainerApply | Sequence[OutputContainerApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(output_container, OutputContainerApply):
            instances = output_container.to_instances_apply()
        else:
            instances = OutputContainerApplyList(output_container).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> OutputContainer:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> OutputContainerList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> OutputContainer | OutputContainerList:
        if isinstance(external_id, str):
            output_container = self._retrieve((self._sources.space, external_id))

            mapping_edges = self.mappings.retrieve(external_id)
            output_container.mappings = [edge.end_node.external_id for edge in mapping_edges]

            return output_container
        else:
            output_containers = self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

            mapping_edges = self.mappings.retrieve(external_id)
            self._set_mappings(output_containers, mapping_edges)

            return output_containers

    def search(
        self,
        query: str,
        properties: OutputContainerTextFields | Sequence[OutputContainerTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> OutputContainerList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            watercourse,
            watercourse_prefix,
            shop_type,
            shop_type_prefix,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _OUTPUTCONTAINER_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: OutputContainerFields | Sequence[OutputContainerFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: OutputContainerTextFields | Sequence[OutputContainerTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: OutputContainerFields | Sequence[OutputContainerFields] | None = None,
        group_by: OutputContainerFields | Sequence[OutputContainerFields] = None,
        query: str | None = None,
        search_properties: OutputContainerTextFields | Sequence[OutputContainerTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: OutputContainerFields | Sequence[OutputContainerFields] | None = None,
        group_by: OutputContainerFields | Sequence[OutputContainerFields] | None = None,
        query: str | None = None,
        search_property: OutputContainerTextFields | Sequence[OutputContainerTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            watercourse,
            watercourse_prefix,
            shop_type,
            shop_type_prefix,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _OUTPUTCONTAINER_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: OutputContainerFields,
        interval: float,
        query: str | None = None,
        search_property: OutputContainerTextFields | Sequence[OutputContainerTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            watercourse,
            watercourse_prefix,
            shop_type,
            shop_type_prefix,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _OUTPUTCONTAINER_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        shop_type: str | list[str] | None = None,
        shop_type_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> OutputContainerList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            watercourse,
            watercourse_prefix,
            shop_type,
            shop_type_prefix,
            external_id_prefix,
            filter,
        )

        output_containers = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            if len(external_ids := output_containers.as_external_ids()) > IN_FILTER_LIMIT:
                mapping_edges = self.mappings.list(limit=-1)
            else:
                mapping_edges = self.mappings.list(external_ids, limit=-1)
            self._set_mappings(output_containers, mapping_edges)

        return output_containers

    @staticmethod
    def _set_mappings(output_containers: Sequence[OutputContainer], mapping_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in mapping_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for output_container in output_containers:
            node_id = output_container.id_tuple()
            if node_id in edges_by_start_node:
                output_container.mappings = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    watercourse: str | list[str] | None = None,
    watercourse_prefix: str | None = None,
    shop_type: str | list[str] | None = None,
    shop_type_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if watercourse and isinstance(watercourse, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("watercourse"), value=watercourse))
    if watercourse and isinstance(watercourse, list):
        filters.append(dm.filters.In(view_id.as_property_ref("watercourse"), values=watercourse))
    if watercourse_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("watercourse"), value=watercourse_prefix))
    if shop_type and isinstance(shop_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopType"), value=shop_type))
    if shop_type and isinstance(shop_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopType"), values=shop_type))
    if shop_type_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopType"), value=shop_type_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from typing import Literal, Optional

import arrow
from cognite.client.data_classes import Event, FileMetadata
from pydantic import BaseModel, ConfigDict, field_validator
from typing_extensions import Self

_SHOP_VERSION_FALLBACK = "15.3.3.2"


class DayaheadFunctionEvent:
    event_type: str = "POWEROPS_FUNCTION_CALL"
    external_id_prefix: str = "POWEROPS_FUNCTION_CALL_"
    function_external_id: str = "function_external_id"
    bid_date: str = "bid:date"
    workflow_event_external_id: str = "workflow_event_external_id"

    @classmethod
    def as_cdf_event(
        cls,
        data_set: int,
        workflow_event_external_id: str,
        bid_date: str,
        event_external_id: str,
        function_name: str,
        additional_metadata: dict,
    ) -> Event:
        return Event(
            external_id=event_external_id,
            type=PartialFunctionEvent.event_type,
            data_set_id=data_set,
            metadata={
                **additional_metadata,
                cls.bid_date: bid_date,
                cls.workflow_event_external_id: workflow_event_external_id,
                cls.function_external_id: function_name,
                "process_type": DayaheadFunctionEvent.event_type,
                "processed": "true",
            },
        )


class PartialFunctionEvent(DayaheadFunctionEvent):
    plant: str = "bid:plant"
    method: str = "bid:bid_matrix_generation_method"
    relationship_label_to_trigger_event: str = "relationship_to.bid_process_event"
    # This is a temporary hack that is introduced due
    # to the potential inconsistency between plant names in
    # SHOP files and the plant asset names in CDF
    plant_name_override: str = "bid:plant_name_override"


class TotalFunctionEvent(DayaheadFunctionEvent):
    portfolio: str = "bid:portfolio"
    bid_process_configuration_name: str = "bid:bid_process_configuration_name"
    relationship_label_to_trigger_event: str = "relationship_to.calculate_total_bid_matrix_event"


class DayaheadTriggerEvent:
    event_type: str = "POWEROPS_BID_PROCESS_FROM_PRERUNS"
    external_id_prefix: str = "POWEROPS_BID_PROCESS_"
    market: str = "bid:market"
    main_scenario: str = "bid:main_scenario"
    market_configuration_nordpool_dayahead: str = "bid:market_config_external_id"
    combine_bid_matrix_tasks: str = "bid:combine_bid_matrix_tasks"
    bid_process_configuration_name: str = "bid:bid_process_configuration_name"
    bid_matrix_generator_config_external_id: str = "bid:bid_matrix_generator_config_external_id"
    price_scenarios: str = "bid:price_scenarios"
    method: str = "bid:bid_process_configuration_name"
    price_area: str = "bid:price_area"
    bid_date: str = "bid:date"
    start_time: str = "shop:starttime"
    end_time: str = "shop:endtime"
    relationship_label_to_shop_run_events: str = "relationship_to.shop_run_event"

    @classmethod
    def as_cdf_event(
        cls,
        event_external_id: str,
        data_set: int,
        start_time: str,
        end_time: str,
        bid_date: str,
        price_scenarios: list[str],
        price_area: str,
        market_configuration_nordpool_dayahead: str,
        bid_configuration_name,
        main_scenario: str = "",
    ) -> Event:
        return Event(
            external_id=event_external_id,
            type=DayaheadTriggerEvent.event_type,
            data_set_id=data_set,
            metadata={
                cls.market: "Dayahead",
                cls.main_scenario: main_scenario,
                cls.price_scenarios: ",".join(price_scenarios),
                cls.bid_date: bid_date,
                cls.start_time: start_time,
                cls.end_time: end_time,
                cls.market_configuration_nordpool_dayahead: market_configuration_nordpool_dayahead,
                cls.bid_process_configuration_name: bid_configuration_name,
                cls.bid_matrix_generator_config_external_id: f"POWEROPS_bid_matrix_generator_config_"
                f"{bid_configuration_name}",
                cls.price_area: f"price_area_{price_area}",
                cls.combine_bid_matrix_tasks: "true",
                "processed": "true",
            },
        )


class PrerunFileMetadata:
    plants: str = "shop:plants"
    price_scenario: str = "shop:price_scenario"
    _plants_delimiter: str = ","


class ShopPreRunFile(BaseModel):
    """
    Represents a single shop run based on a pre-run file.
    """

    pre_run_external_id: str
    plants: list[str]
    price_scenario: str

    @classmethod
    def load_from_metadata(cls, file_metadata: FileMetadata) -> Self:
        plants_as_string = file_metadata.metadata.get(PrerunFileMetadata.plants)
        plants = [plant.lstrip() for plant in plants_as_string.split(PrerunFileMetadata._plants_delimiter)]
        return cls(
            pre_run_external_id=file_metadata.external_id,
            plants=plants,
            price_scenario=file_metadata.metadata.get(PrerunFileMetadata.price_scenario),
        )


class Case(BaseModel):
    """
    Case definition based on a set of prerun files to run for certain plants that are used in the set of prerun files
    """

    case_name: str  # e.g. Glomma
    pre_run_file_external_ids: list[str]
    commands_file: Optional[str] = None
    pre_runs_external_id_prefix: Optional[str] = None


class BidTimeFrame(BaseModel):
    """
    Used to dynamically specify what times to run the Dayahead bid process for in local time
    :shift_start_in_days: number of days from the current local time to shift the start time of the bid
    :timezone: the local timezone to use when creating shop times
    :bid_period_in_days: the number of days the bid is valid for. This number is used to produce the bid end time by
    shifting the start time with this number
    """

    model_config = ConfigDict(frozen=True)

    shift_start_in_days: int
    timezone: Optional[str] = "Europe/Oslo"
    bid_period_in_days: Optional[int] = 14
    _datetime_string_format: str = "YYYY-MM-DD HH:mm:ss"
    _date_string_format: str = "YYYY-MM-DD"
    _start_time_arrow: arrow.Arrow

    @field_validator("bid_period_in_days")
    @classmethod
    def positive_bid_period_in_days(cls, value):
        if value < 0:
            raise ValueError("Bid period must be a positive number")

    @property
    def start_time_arrow(self):
        return (
            arrow.now(self.timezone).shift(days=self.shift_start_in_days).floor("day")
        )  # should floor be hour instead?

    @property
    def start_time_string(self) -> str:
        return self.start_time_arrow.format(self._datetime_string_format)

    @property
    def end_time_string(self) -> str:
        return (
            self.start_time_arrow.shift(days=self.bid_period_in_days).floor("week").format(self._datetime_string_format)
        )

    @property
    def bid_date(self) -> str:
        return self.start_time_arrow.shift(days=1).format(self._date_string_format)


class DayaheadTrigger(BaseModel):
    """
    Contains the blueprint of a Dayahead workflow definition.
    This object is used together with the DayaheadTriggerAPI to trigger a workflow from an instance of this class.
    """

    price_scenarios: list[str]
    main_scenario: str = ""
    shop_version: Optional[str] = _SHOP_VERSION_FALLBACK
    price_area: str
    method: Literal["multi_scenario", "price_independent"]
    bid_configuration_name: str
    cases: list[Case]
    bid_time_frame: Optional[BidTimeFrame] = BidTimeFrame(shift_start_in_days=0)
    plant_names_override: Optional[dict] = None
    dayahead_configuration_external_id: Optional[str] = "market_configuration_nordpool_dayahead"


class DayaheadWorkflowRun(BaseModel):
    """
    Return object when triggering a DayaheadTrigger workflow via the API.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    workflow_trigger_event: Event
    total_bid_event: Event
    partial_bid_events: list[Event]
    shop_run_events: list[Event]



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._time_interval import TimeIntervalApply

__all__ = [
    "BidDocumentHeader",
    "BidDocumentHeaderApply",
    "BidDocumentHeaderList",
    "BidDocumentHeaderApplyList",
    "BidDocumentHeaderFields",
    "BidDocumentHeaderTextFields",
]


BidDocumentHeaderTextFields = Literal["document_type_name", "process_type_name", "tso", "owner", "country", "origin"]
BidDocumentHeaderFields = Literal[
    "document_type_name", "process_type_name", "tso", "owner", "created_date_time", "country", "origin"
]

_BIDDOCUMENTHEADER_PROPERTIES_BY_FIELD = {
    "document_type_name": "DocumentTypeName",
    "process_type_name": "ProcessTypeName",
    "tso": "TSO",
    "owner": "Owner",
    "created_date_time": "CreatedDateTime",
    "country": "Country",
    "origin": "Origin",
}


class BidDocumentHeader(DomainModel):
    space: str = "power-ops"
    document_type_name: Optional[str] = Field(None, alias="DocumentTypeName")
    process_type_name: Optional[str] = Field(None, alias="ProcessTypeName")
    tso: Optional[str] = Field(None, alias="TSO")
    owner: Optional[str] = Field(None, alias="Owner")
    created_date_time: Optional[datetime.datetime] = Field(None, alias="CreatedDateTime")
    bid_interval: Optional[str] = Field(None, alias="BidInterval")
    country: Optional[str] = Field(None, alias="Country")
    origin: Optional[str] = Field(None, alias="Origin")

    def as_apply(self) -> BidDocumentHeaderApply:
        return BidDocumentHeaderApply(
            external_id=self.external_id,
            document_type_name=self.document_type_name,
            process_type_name=self.process_type_name,
            tso=self.tso,
            owner=self.owner,
            created_date_time=self.created_date_time,
            bid_interval=self.bid_interval,
            country=self.country,
            origin=self.origin,
        )


class BidDocumentHeaderApply(DomainModelApply):
    space: str = "power-ops"
    document_type_name: Optional[str] = Field(None, alias="DocumentTypeName")
    process_type_name: Optional[str] = Field(None, alias="ProcessTypeName")
    tso: Optional[str] = Field(None, alias="TSO")
    owner: Optional[str] = Field(None, alias="Owner")
    created_date_time: Optional[datetime.datetime] = Field(None, alias="CreatedDateTime")
    bid_interval: Union[TimeIntervalApply, str, None] = Field(None, repr=False, alias="BidInterval")
    country: Optional[str] = Field(None, alias="Country")
    origin: Optional[str] = Field(None, alias="Origin")

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.document_type_name is not None:
            properties["DocumentTypeName"] = self.document_type_name
        if self.process_type_name is not None:
            properties["ProcessTypeName"] = self.process_type_name
        if self.tso is not None:
            properties["TSO"] = self.tso
        if self.owner is not None:
            properties["Owner"] = self.owner
        if self.created_date_time is not None:
            properties["CreatedDateTime"] = self.created_date_time.isoformat(timespec="milliseconds")
        if self.bid_interval is not None:
            properties["BidInterval"] = {
                "space": "power-ops",
                "externalId": self.bid_interval
                if isinstance(self.bid_interval, str)
                else self.bid_interval.external_id,
            }
        if self.country is not None:
            properties["Country"] = self.country
        if self.origin is not None:
            properties["Origin"] = self.origin
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "BidDocumentHeader"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        if isinstance(self.bid_interval, DomainModelApply):
            instances = self.bid_interval._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class BidDocumentHeaderList(TypeList[BidDocumentHeader]):
    _NODE = BidDocumentHeader

    def as_apply(self) -> BidDocumentHeaderApplyList:
        return BidDocumentHeaderApplyList([node.as_apply() for node in self.data])


class BidDocumentHeaderApplyList(TypeApplyList[BidDocumentHeaderApply]):
    _NODE = BidDocumentHeaderApply



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = ["Reason", "ReasonApply", "ReasonList", "ReasonApplyList", "ReasonFields", "ReasonTextFields"]


ReasonTextFields = Literal["code", "text"]
ReasonFields = Literal["code", "text"]

_REASON_PROPERTIES_BY_FIELD = {
    "code": "code",
    "text": "text",
}


class Reason(DomainModel):
    space: str = "power-ops"
    code: Optional[str] = None
    text: Optional[str] = None

    def as_apply(self) -> ReasonApply:
        return ReasonApply(
            external_id=self.external_id,
            code=self.code,
            text=self.text,
        )


class ReasonApply(DomainModelApply):
    space: str = "power-ops"
    code: Optional[str] = None
    text: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.code is not None:
            properties["code"] = self.code
        if self.text is not None:
            properties["text"] = self.text
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "Reason"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class ReasonList(TypeList[Reason]):
    _NODE = Reason

    def as_apply(self) -> ReasonApplyList:
        return ReasonApplyList([node.as_apply() for node in self.data])


class ReasonApplyList(TypeApplyList[ReasonApply]):
    _NODE = ReasonApply



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._date_transformation import DateTransformationApply
    from ._reserve_scenario import ReserveScenarioApply
    from ._rkom_market import RKOMMarketApply
    from ._scenario_mapping import ScenarioMappingApply

__all__ = ["RKOMBid", "RKOMBidApply", "RKOMBidList", "RKOMBidApplyList", "RKOMBidFields", "RKOMBidTextFields"]


RKOMBidTextFields = Literal["name", "method", "watercourse"]
RKOMBidFields = Literal["name", "method", "minimum_price", "price_premium", "watercourse"]

_RKOMBID_PROPERTIES_BY_FIELD = {
    "name": "name",
    "method": "method",
    "minimum_price": "minimumPrice",
    "price_premium": "pricePremium",
    "watercourse": "watercourse",
}


class RKOMBid(DomainModel):
    space: str = "power-ops"
    name: Optional[str] = None
    market: Optional[str] = None
    method: Optional[str] = None
    minimum_price: Optional[float] = Field(None, alias="minimumPrice")
    price_premium: Optional[float] = Field(None, alias="pricePremium")
    watercourse: Optional[str] = None
    date: Optional[list[str]] = None
    price_scenarios: Optional[list[str]] = Field(None, alias="priceScenarios")
    reserve_scenarios: Optional[list[str]] = Field(None, alias="reserveScenarios")

    def as_apply(self) -> RKOMBidApply:
        return RKOMBidApply(
            external_id=self.external_id,
            name=self.name,
            market=self.market,
            method=self.method,
            minimum_price=self.minimum_price,
            price_premium=self.price_premium,
            watercourse=self.watercourse,
            date=self.date,
            price_scenarios=self.price_scenarios,
            reserve_scenarios=self.reserve_scenarios,
        )


class RKOMBidApply(DomainModelApply):
    space: str = "power-ops"
    name: Optional[str] = None
    market: Union[RKOMMarketApply, str, None] = Field(None, repr=False)
    method: Optional[str] = None
    minimum_price: Optional[float] = Field(None, alias="minimumPrice")
    price_premium: Optional[float] = Field(None, alias="pricePremium")
    watercourse: Optional[str] = None
    date: Union[list[DateTransformationApply], list[str], None] = Field(default=None, repr=False)
    price_scenarios: Union[list[ScenarioMappingApply], list[str], None] = Field(
        default=None, repr=False, alias="priceScenarios"
    )
    reserve_scenarios: Union[list[ReserveScenarioApply], list[str], None] = Field(
        default=None, repr=False, alias="reserveScenarios"
    )

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.name is not None:
            properties["name"] = self.name
        if self.market is not None:
            properties["market"] = {
                "space": "power-ops",
                "externalId": self.market if isinstance(self.market, str) else self.market.external_id,
            }
        if self.method is not None:
            properties["method"] = self.method
        if self.minimum_price is not None:
            properties["minimumPrice"] = self.minimum_price
        if self.price_premium is not None:
            properties["pricePremium"] = self.price_premium
        if self.watercourse is not None:
            properties["watercourse"] = self.watercourse
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "RKOMBid"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for date in self.date or []:
            edge = self._create_date_edge(date)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(date, DomainModelApply):
                instances = date._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for price_scenario in self.price_scenarios or []:
            edge = self._create_price_scenario_edge(price_scenario)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(price_scenario, DomainModelApply):
                instances = price_scenario._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for reserve_scenario in self.reserve_scenarios or []:
            edge = self._create_reserve_scenario_edge(reserve_scenario)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(reserve_scenario, DomainModelApply):
                instances = reserve_scenario._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.market, DomainModelApply):
            instances = self.market._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_date_edge(self, date: Union[str, DateTransformationApply]) -> dm.EdgeApply:
        if isinstance(date, str):
            end_node_ext_id = date
        elif isinstance(date, DomainModelApply):
            end_node_ext_id = date.external_id
        else:
            raise TypeError(f"Expected str or DateTransformationApply, got {type(date)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "RKOMBid.date"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )

    def _create_price_scenario_edge(self, price_scenario: Union[str, ScenarioMappingApply]) -> dm.EdgeApply:
        if isinstance(price_scenario, str):
            end_node_ext_id = price_scenario
        elif isinstance(price_scenario, DomainModelApply):
            end_node_ext_id = price_scenario.external_id
        else:
            raise TypeError(f"Expected str or ScenarioMappingApply, got {type(price_scenario)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "RKOMBid.priceScenarios"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )

    def _create_reserve_scenario_edge(self, reserve_scenario: Union[str, ReserveScenarioApply]) -> dm.EdgeApply:
        if isinstance(reserve_scenario, str):
            end_node_ext_id = reserve_scenario
        elif isinstance(reserve_scenario, DomainModelApply):
            end_node_ext_id = reserve_scenario.external_id
        else:
            raise TypeError(f"Expected str or ReserveScenarioApply, got {type(reserve_scenario)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "RKOMBid.reserveScenarios"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class RKOMBidList(TypeList[RKOMBid]):
    _NODE = RKOMBid

    def as_apply(self) -> RKOMBidApplyList:
        return RKOMBidApplyList([node.as_apply() for node in self.data])


class RKOMBidApplyList(TypeApplyList[RKOMBidApply]):
    _NODE = RKOMBidApply



"""
Checks whether time series referenced in a model actually exist in CDF
and have data in the required time ranges.
"""

from __future__ import annotations

import json
import logging
from collections import defaultdict
from collections.abc import Sequence
from enum import Enum
from typing import Literal, Optional, Union, cast

import arrow
import numpy as np
import pandas as pd
from pydantic import BaseModel, ConfigDict, field_serializer, model_validator

from cognite.powerops import PowerOpsClient
from cognite.powerops.client._generated.data_classes import (
    InputTimeSeriesMapping,
    InputTimeSeriesMappingApply,
)
from cognite.powerops.resync.models.base import Model
from cognite.powerops.utils.cdf.calls import retrieve_time_series_datapoints
from cognite.powerops.utils.lookup import attr_lookup, dict_values, each
from cognite.powerops.utils.require import require
from cognite.powerops.utils.time import arrow_to_ms, relative_time_specification_to_arrow

logger = logging.getLogger(__name__)


class ValidationSpec(BaseModel):
    pass


class TimeSeriesValidation(ValidationSpec):
    mapping: Union[InputTimeSeriesMapping, InputTimeSeriesMappingApply]
    validation_range: ValidationRange
    data_models: list[str] = []

    model_config = ConfigDict(arbitrary_types_allowed=True)


class ValidationRange(BaseModel):
    start: arrow.Arrow
    end: arrow.Arrow

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __str__(self):
        return f"{self.start} - {self.end}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ValidationRange):
            return self.start == other.start and self.end == other.end
        return super().__eq__(other)

    @field_serializer("start", "end")
    def serialize_arrow(self, value: arrow.Arrow, _info):
        return value.isoformat()


class TimeSeriesValidationFailures(str, Enum):
    NO_DATAPOINTS = "NO_DATAPOINTS"
    SINGLE_DATAPOINT = "SINGLE_DATAPOINT"


class ValidationResult(BaseModel):
    valid: bool = False


class ValidationResults(BaseModel):
    results: Sequence[ValidationResult]

    def as_markdown(self, include_valid: bool) -> str:
        raise NotImplementedError()

    def as_json(self, include_valid: bool) -> str:
        results = self.results
        if not include_valid:
            results = [result for result in results if not result.valid]
        return json.dumps([result.model_dump(mode="json") for result in results], ensure_ascii=False)


class TimeSeriesValidationResults(ValidationResults):
    results: Sequence[TimeSeriesValidationResult]

    def as_markdown(self, include_valid: bool) -> str:
        lines = [
            "| Time Range                                            | Valid | models | reason | ExternalID |",
            "|-------------------------------------------------------|-------|--------|--------|------------|",
        ]
        for result in self.results:
            if result.valid and not include_valid:
                continue
            time_range = str(result.validation.validation_range)
            valid = str(result.valid)
            models = ",".join(result.validation.data_models)
            reason = result.reason
            external_id = result.validation.mapping.cdf_time_series
            lines.append(f"| {time_range} | {valid:5} | {models} | {reason} | {external_id} |")
        return "\n".join(lines)


class TimeSeriesValidationResult(ValidationResult):
    failure: Optional[TimeSeriesValidationFailures] = None
    validation: TimeSeriesValidation

    @model_validator(mode="before")
    @classmethod
    def set_valid(cls, data: dict) -> dict:
        if "valid" not in data and "failure" not in data:
            raise ValueError("One of valid=True or failure is required.")
        if data.get("valid") and "failure" in data:
            raise ValueError("One of valid=True or failure is required.")
        if "valid" not in data:
            data["valid"] = not bool(data.get("failure"))
        return data

    @property
    def reason(self) -> str:
        return "" if self.valid else require(self.failure).value


def find_mappings(obj, lookup: Literal["base", "process"]) -> list[InputTimeSeriesMappingApply]:
    def df_to_dict_records(item, attr_path):
        yield from attr_lookup(item.replace(np.nan, None).to_dict("records"), attr_path)

    def to_input_time_series_mapping_apply(item, _attr_path):
        if item.get("time_series_external_id") is not None:
            shop_obj_type, shop_obj_name, shop_attr_name = item["shop_model_path"].split(".")
            yield InputTimeSeriesMappingApply(
                external_id=item["time_series_external_id"],
                aggregation=item["aggregation"],
                retrieve=item["retrieve"],
                cdf_time_series=item["time_series_external_id"],
                transformations=[],  # TODO fill out from item["transformations"] + item["transformations1"] + ...
                shop_object_type=shop_obj_type,
                shop_object_name=shop_obj_name,
                shop_attribute_name=shop_attr_name,
            )

    def from_mapping_to_input_time_series_mapping_apply(item, _attr_path):
        if item.timeseries_external_id is not None:
            shop_obj_type, shop_obj_name, shop_attr_name = item.path.split(".")
            yield InputTimeSeriesMappingApply(
                external_id=item.timeseries_external_id,
                aggregation=item.aggregation,
                retrieve=item.retrieve,
                cdf_time_series=item.timeseries_external_id,
                transformations=[],  # TODO fill out rom item.transformations
                shop_object_type=shop_obj_type,
                shop_object_name=shop_obj_name,
                shop_attribute_name=shop_attr_name,
            )

    lookup_paths = {
        "process": [
            ["incremental_mapping", each, "content", df_to_dict_records, each, to_input_time_series_mapping_apply],
        ],
        "base": [
            ["mappings", dict_values, each, from_mapping_to_input_time_series_mapping_apply],
            #  ^ includes scenario mappings
        ],
    }

    mappings: list[InputTimeSeriesMappingApply] = []
    for lookup_path in lookup_paths[lookup]:
        mappings.extend(filter(None, list(attr_lookup(obj, lookup_path))))
    return mappings


PreparedValidationsT = dict[str, dict[str, TimeSeriesValidation]]
ValidationRangesT = dict[str, ValidationRange]


def prepare_validation(models: list[Model]) -> tuple[PreparedValidationsT, ValidationRangesT]:
    # data containers:
    ts_validations = cast(PreparedValidationsT, defaultdict(lambda: defaultdict(dict)))
    validation_ranges: dict[str, ValidationRange] = {}

    # Find processes (i.e. configuration for actual SHOP runs, with start and end times)
    # Without this we don't know what time ranges to query from CDF.
    for model in models:
        if not hasattr(model, "processes"):
            continue
        for process in model.processes:
            start_time = relative_time_specification_to_arrow(process.shop.starttime)
            end_time = relative_time_specification_to_arrow(process.shop.endtime)
            validation_range = ValidationRange(start=start_time, end=end_time)
            validation_ranges[str(validation_range)] = validation_range

            # Find mappings that are configured within processes.
            # For these mapping we only validate time ranges of the process.
            process_mappings = find_mappings(process, "process")
            for mapping in process_mappings:
                ts_validation = ts_validations[f"{validation_range}"].get(mapping.external_id)
                if ts_validation is None:
                    ts_validation = TimeSeriesValidation(
                        mapping=mapping,
                        validation_range=validation_range,
                    )
                if model.model_name not in ts_validation.data_models:
                    ts_validation.data_models.append(model.model_name)
                ts_validations[f"{validation_range}"][mapping.external_id] = ts_validation

    # Find base mappings.
    # For these mappings we don't know the time ranges, so we validate all time ranges.
    # Overwrite any process-mappings from above.
    for model in models:
        base_mappings = find_mappings(model, "base")
        for mapping in base_mappings:
            for validation_range_str, validation_range in validation_ranges.items():
                ts_validation = TimeSeriesValidation(
                    mapping=mapping,
                    validation_range=validation_range,
                )
                if model.model_name not in ts_validation.data_models:
                    ts_validation.data_models.append(model.model_name)
                ts_validations[validation_range_str][mapping.external_id] = ts_validation

    # Get any additional timeseries that are part of the models but are not part of the mappings.
    # These are ambiguous... TODO Overwrite previous mappings or not? These are hard to connect back to SHOP path.
    additional_timeseries = [ts for model in models for ts in model.timeseries()]
    for additional_ts in additional_timeseries:
        for validation_range_str in validation_ranges:
            ts_validations[validation_range_str][require(additional_ts.external_id)] = TimeSeriesValidation(
                validation_range=validation_ranges[validation_range_str],
                mapping=InputTimeSeriesMappingApply(  # dummy mapping, for validation only
                    external_id=require(additional_ts.external_id),  # type: ignore[call-arg]
                    cdf_time_series=require(additional_ts.external_id),  # type: ignore[call-arg]
                    retrieve="RANGE",
                    shop_attribute_name="N/A",  # could get this from the model, possibly enhance .timeseries()
                ),
            )

    return ts_validations, validation_ranges


def perform_validation(
    po_client: PowerOpsClient, ts_validations: PreparedValidationsT, validation_ranges: ValidationRangesT
) -> ValidationResults:
    results: list[TimeSeriesValidationResult] = []
    for range_str, validations_in_range in ts_validations.items():
        validation_range = validation_ranges[range_str]
        all_mappings = [validation.mapping for validation in validations_in_range.values()]
        ts_mappings = list(filter(lambda row: row.cdf_time_series, all_mappings))

        logger.info(f"Retrieving datapoints for {len(ts_mappings)} mappings from range {range_str}")
        datapoints = retrieve_time_series_datapoints(
            po_client.cdf,
            ts_mappings,
            start=arrow_to_ms(validation_range.start),
            end=arrow_to_ms(validation_range.end),
        )
        warnings_n = 0
        for ts_mapping in ts_mappings:
            data: pd.Series = datapoints.get(require(ts_mapping.cdf_time_series), pd.Series())
            ts_validation: TimeSeriesValidation = ts_validations[range_str][ts_mapping.external_id]
            if not len(data):
                logger.info(
                    f"VALIDATION FAIL: No datapoints found for range {range_str} in timeseries "
                    f"'{ts_mapping.cdf_time_series}', used in models: {', '.join(ts_validation.data_models)}.",
                )
                results.append(
                    TimeSeriesValidationResult(
                        validation=ts_validation, failure=TimeSeriesValidationFailures.NO_DATAPOINTS
                    )
                )
                warnings_n += 1
            elif ts_mapping.retrieve == "RANGE" and len(data) < 2:
                logger.info(
                    f"VALIDATION FAIL: Only one datapoint found for range {range_str} in timeseries "
                    f"'{ts_mapping.cdf_time_series}', used in models: {', '.join(ts_validation.data_models)}.",
                )
                results.append(
                    TimeSeriesValidationResult(
                        validation=ts_validation, failure=TimeSeriesValidationFailures.SINGLE_DATAPOINT
                    )
                )
                warnings_n += 1
            else:
                results.append(TimeSeriesValidationResult(validation=ts_validation, valid=True))
        if warnings_n:
            logger.info(f"Total warnings for range {range_str}: {warnings_n}.")
    return TimeSeriesValidationResults(results=results)



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.data_classes import (
    BidDocumentHeader,
    BidDocumentHeaderApply,
    BidDocumentHeaderApplyList,
    BidDocumentHeaderFields,
    BidDocumentHeaderList,
    BidDocumentHeaderTextFields,
)
from cognite.powerops.client._generated.data_classes._bid_document_header import _BIDDOCUMENTHEADER_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, Aggregations, TypeAPI


class BidDocumentHeaderAPI(TypeAPI[BidDocumentHeader, BidDocumentHeaderApply, BidDocumentHeaderList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=BidDocumentHeader,
            class_apply_type=BidDocumentHeaderApply,
            class_list=BidDocumentHeaderList,
        )
        self._view_id = view_id

    def apply(
        self, bid_document_header: BidDocumentHeaderApply | Sequence[BidDocumentHeaderApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(bid_document_header, BidDocumentHeaderApply):
            instances = bid_document_header.to_instances_apply()
        else:
            instances = BidDocumentHeaderApplyList(bid_document_header).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> BidDocumentHeader:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> BidDocumentHeaderList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> BidDocumentHeader | BidDocumentHeaderList:
        if isinstance(external_id, str):
            return self._retrieve((self._sources.space, external_id))
        else:
            return self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

    def search(
        self,
        query: str,
        properties: BidDocumentHeaderTextFields | Sequence[BidDocumentHeaderTextFields] | None = None,
        document_type_name: str | list[str] | None = None,
        document_type_name_prefix: str | None = None,
        process_type_name: str | list[str] | None = None,
        process_type_name_prefix: str | None = None,
        tso: str | list[str] | None = None,
        tso_prefix: str | None = None,
        owner: str | list[str] | None = None,
        owner_prefix: str | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        bid_interval: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        country: str | list[str] | None = None,
        country_prefix: str | None = None,
        origin: str | list[str] | None = None,
        origin_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> BidDocumentHeaderList:
        filter_ = _create_filter(
            self._view_id,
            document_type_name,
            document_type_name_prefix,
            process_type_name,
            process_type_name_prefix,
            tso,
            tso_prefix,
            owner,
            owner_prefix,
            min_created_date_time,
            max_created_date_time,
            bid_interval,
            country,
            country_prefix,
            origin,
            origin_prefix,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _BIDDOCUMENTHEADER_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidDocumentHeaderFields | Sequence[BidDocumentHeaderFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: BidDocumentHeaderTextFields | Sequence[BidDocumentHeaderTextFields] | None = None,
        document_type_name: str | list[str] | None = None,
        document_type_name_prefix: str | None = None,
        process_type_name: str | list[str] | None = None,
        process_type_name_prefix: str | None = None,
        tso: str | list[str] | None = None,
        tso_prefix: str | None = None,
        owner: str | list[str] | None = None,
        owner_prefix: str | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        bid_interval: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        country: str | list[str] | None = None,
        country_prefix: str | None = None,
        origin: str | list[str] | None = None,
        origin_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidDocumentHeaderFields | Sequence[BidDocumentHeaderFields] | None = None,
        group_by: BidDocumentHeaderFields | Sequence[BidDocumentHeaderFields] = None,
        query: str | None = None,
        search_properties: BidDocumentHeaderTextFields | Sequence[BidDocumentHeaderTextFields] | None = None,
        document_type_name: str | list[str] | None = None,
        document_type_name_prefix: str | None = None,
        process_type_name: str | list[str] | None = None,
        process_type_name_prefix: str | None = None,
        tso: str | list[str] | None = None,
        tso_prefix: str | None = None,
        owner: str | list[str] | None = None,
        owner_prefix: str | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        bid_interval: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        country: str | list[str] | None = None,
        country_prefix: str | None = None,
        origin: str | list[str] | None = None,
        origin_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidDocumentHeaderFields | Sequence[BidDocumentHeaderFields] | None = None,
        group_by: BidDocumentHeaderFields | Sequence[BidDocumentHeaderFields] | None = None,
        query: str | None = None,
        search_property: BidDocumentHeaderTextFields | Sequence[BidDocumentHeaderTextFields] | None = None,
        document_type_name: str | list[str] | None = None,
        document_type_name_prefix: str | None = None,
        process_type_name: str | list[str] | None = None,
        process_type_name_prefix: str | None = None,
        tso: str | list[str] | None = None,
        tso_prefix: str | None = None,
        owner: str | list[str] | None = None,
        owner_prefix: str | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        bid_interval: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        country: str | list[str] | None = None,
        country_prefix: str | None = None,
        origin: str | list[str] | None = None,
        origin_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            document_type_name,
            document_type_name_prefix,
            process_type_name,
            process_type_name_prefix,
            tso,
            tso_prefix,
            owner,
            owner_prefix,
            min_created_date_time,
            max_created_date_time,
            bid_interval,
            country,
            country_prefix,
            origin,
            origin_prefix,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _BIDDOCUMENTHEADER_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: BidDocumentHeaderFields,
        interval: float,
        query: str | None = None,
        search_property: BidDocumentHeaderTextFields | Sequence[BidDocumentHeaderTextFields] | None = None,
        document_type_name: str | list[str] | None = None,
        document_type_name_prefix: str | None = None,
        process_type_name: str | list[str] | None = None,
        process_type_name_prefix: str | None = None,
        tso: str | list[str] | None = None,
        tso_prefix: str | None = None,
        owner: str | list[str] | None = None,
        owner_prefix: str | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        bid_interval: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        country: str | list[str] | None = None,
        country_prefix: str | None = None,
        origin: str | list[str] | None = None,
        origin_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            document_type_name,
            document_type_name_prefix,
            process_type_name,
            process_type_name_prefix,
            tso,
            tso_prefix,
            owner,
            owner_prefix,
            min_created_date_time,
            max_created_date_time,
            bid_interval,
            country,
            country_prefix,
            origin,
            origin_prefix,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _BIDDOCUMENTHEADER_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        document_type_name: str | list[str] | None = None,
        document_type_name_prefix: str | None = None,
        process_type_name: str | list[str] | None = None,
        process_type_name_prefix: str | None = None,
        tso: str | list[str] | None = None,
        tso_prefix: str | None = None,
        owner: str | list[str] | None = None,
        owner_prefix: str | None = None,
        min_created_date_time: datetime.datetime | None = None,
        max_created_date_time: datetime.datetime | None = None,
        bid_interval: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        country: str | list[str] | None = None,
        country_prefix: str | None = None,
        origin: str | list[str] | None = None,
        origin_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> BidDocumentHeaderList:
        filter_ = _create_filter(
            self._view_id,
            document_type_name,
            document_type_name_prefix,
            process_type_name,
            process_type_name_prefix,
            tso,
            tso_prefix,
            owner,
            owner_prefix,
            min_created_date_time,
            max_created_date_time,
            bid_interval,
            country,
            country_prefix,
            origin,
            origin_prefix,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    document_type_name: str | list[str] | None = None,
    document_type_name_prefix: str | None = None,
    process_type_name: str | list[str] | None = None,
    process_type_name_prefix: str | None = None,
    tso: str | list[str] | None = None,
    tso_prefix: str | None = None,
    owner: str | list[str] | None = None,
    owner_prefix: str | None = None,
    min_created_date_time: datetime.datetime | None = None,
    max_created_date_time: datetime.datetime | None = None,
    bid_interval: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    country: str | list[str] | None = None,
    country_prefix: str | None = None,
    origin: str | list[str] | None = None,
    origin_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if document_type_name and isinstance(document_type_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("DocumentTypeName"), value=document_type_name))
    if document_type_name and isinstance(document_type_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("DocumentTypeName"), values=document_type_name))
    if document_type_name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("DocumentTypeName"), value=document_type_name_prefix))
    if process_type_name and isinstance(process_type_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("ProcessTypeName"), value=process_type_name))
    if process_type_name and isinstance(process_type_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("ProcessTypeName"), values=process_type_name))
    if process_type_name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("ProcessTypeName"), value=process_type_name_prefix))
    if tso and isinstance(tso, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("TSO"), value=tso))
    if tso and isinstance(tso, list):
        filters.append(dm.filters.In(view_id.as_property_ref("TSO"), values=tso))
    if tso_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("TSO"), value=tso_prefix))
    if owner and isinstance(owner, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("Owner"), value=owner))
    if owner and isinstance(owner, list):
        filters.append(dm.filters.In(view_id.as_property_ref("Owner"), values=owner))
    if owner_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("Owner"), value=owner_prefix))
    if min_created_date_time or max_created_date_time:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("CreatedDateTime"),
                gte=min_created_date_time.isoformat() if min_created_date_time else None,
                lte=max_created_date_time.isoformat() if max_created_date_time else None,
            )
        )
    if bid_interval and isinstance(bid_interval, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("BidInterval"), value={"space": "power-ops", "externalId": bid_interval}
            )
        )
    if bid_interval and isinstance(bid_interval, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("BidInterval"), value={"space": bid_interval[0], "externalId": bid_interval[1]}
            )
        )
    if bid_interval and isinstance(bid_interval, list) and isinstance(bid_interval[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("BidInterval"),
                values=[{"space": "power-ops", "externalId": item} for item in bid_interval],
            )
        )
    if bid_interval and isinstance(bid_interval, list) and isinstance(bid_interval[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("BidInterval"),
                values=[{"space": item[0], "externalId": item[1]} for item in bid_interval],
            )
        )
    if country and isinstance(country, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("Country"), value=country))
    if country and isinstance(country, list):
        filters.append(dm.filters.In(view_id.as_property_ref("Country"), values=country))
    if country_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("Country"), value=country_prefix))
    if origin and isinstance(origin, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("Origin"), value=origin))
    if origin and isinstance(origin, list):
        filters.append(dm.filters.In(view_id.as_property_ref("Origin"), values=origin))
    if origin_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("Origin"), value=origin_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from ._benchmark_bid import (
    BenchmarkBid,
    BenchmarkBidApply,
    BenchmarkBidApplyList,
    BenchmarkBidFields,
    BenchmarkBidList,
    BenchmarkBidTextFields,
)
from ._benchmark_process import (
    BenchmarkProcess,
    BenchmarkProcessApply,
    BenchmarkProcessApplyList,
    BenchmarkProcessFields,
    BenchmarkProcessList,
    BenchmarkProcessTextFields,
)
from ._bid_curves import (
    BidCurves,
    BidCurvesApply,
    BidCurvesApplyList,
    BidCurvesFields,
    BidCurvesList,
    BidCurvesTextFields,
)
from ._bid_document_header import (
    BidDocumentHeader,
    BidDocumentHeaderApply,
    BidDocumentHeaderApplyList,
    BidDocumentHeaderFields,
    BidDocumentHeaderList,
    BidDocumentHeaderTextFields,
)
from ._bid_matrix_generator import (
    BidMatrixGenerator,
    BidMatrixGeneratorApply,
    BidMatrixGeneratorApplyList,
    BidMatrixGeneratorFields,
    BidMatrixGeneratorList,
    BidMatrixGeneratorTextFields,
)
from ._bid_time_series import (
    BidTimeSeries,
    BidTimeSeriesApply,
    BidTimeSeriesApplyList,
    BidTimeSeriesFields,
    BidTimeSeriesList,
    BidTimeSeriesTextFields,
)
from ._command_config import (
    CommandConfig,
    CommandConfigApply,
    CommandConfigApplyList,
    CommandConfigFields,
    CommandConfigList,
    CommandConfigTextFields,
)
from ._core import DomainModel, DomainModelApply
from ._date_time_interval import (
    DateTimeInterval,
    DateTimeIntervalApply,
    DateTimeIntervalApplyList,
    DateTimeIntervalFields,
    DateTimeIntervalList,
)
from ._date_transformation import (
    DateTransformation,
    DateTransformationApply,
    DateTransformationApplyList,
    DateTransformationFields,
    DateTransformationList,
    DateTransformationTextFields,
)
from ._day_ahead_bid import (
    DayAheadBid,
    DayAheadBidApply,
    DayAheadBidApplyList,
    DayAheadBidFields,
    DayAheadBidList,
    DayAheadBidTextFields,
)
from ._day_ahead_process import (
    DayAheadProcess,
    DayAheadProcessApply,
    DayAheadProcessApplyList,
    DayAheadProcessFields,
    DayAheadProcessList,
    DayAheadProcessTextFields,
)
from ._duration import Duration, DurationApply, DurationApplyList, DurationFields, DurationList, DurationTextFields
from ._generator import (
    Generator,
    GeneratorApply,
    GeneratorApplyList,
    GeneratorFields,
    GeneratorList,
    GeneratorTextFields,
)
from ._input_time_series_mapping import (
    InputTimeSeriesMapping,
    InputTimeSeriesMappingApply,
    InputTimeSeriesMappingApplyList,
    InputTimeSeriesMappingFields,
    InputTimeSeriesMappingList,
    InputTimeSeriesMappingTextFields,
)
from ._market_agreement import (
    MarketAgreement,
    MarketAgreementApply,
    MarketAgreementApplyList,
    MarketAgreementFields,
    MarketAgreementList,
    MarketAgreementTextFields,
)
from ._market_participant import (
    MarketParticipant,
    MarketParticipantApply,
    MarketParticipantApplyList,
    MarketParticipantFields,
    MarketParticipantList,
    MarketParticipantTextFields,
)
from ._mba_domain import (
    MBADomain,
    MBADomainApply,
    MBADomainApplyList,
    MBADomainFields,
    MBADomainList,
    MBADomainTextFields,
)
from ._nord_pool_market import (
    NordPoolMarket,
    NordPoolMarketApply,
    NordPoolMarketApplyList,
    NordPoolMarketFields,
    NordPoolMarketList,
    NordPoolMarketTextFields,
)
from ._output_container import (
    OutputContainer,
    OutputContainerApply,
    OutputContainerApplyList,
    OutputContainerFields,
    OutputContainerList,
    OutputContainerTextFields,
)
from ._output_mapping import (
    OutputMapping,
    OutputMappingApply,
    OutputMappingApplyList,
    OutputMappingFields,
    OutputMappingList,
    OutputMappingTextFields,
)
from ._periods import Periods, PeriodsApply, PeriodsApplyList, PeriodsFields, PeriodsList, PeriodsTextFields
from ._plant import Plant, PlantApply, PlantApplyList, PlantFields, PlantList, PlantTextFields
from ._point import Point, PointApply, PointApplyList, PointFields, PointList
from ._price_area import (
    PriceArea,
    PriceAreaApply,
    PriceAreaApplyList,
    PriceAreaFields,
    PriceAreaList,
    PriceAreaTextFields,
)
from ._production_plan_time_series import (
    ProductionPlanTimeSeries,
    ProductionPlanTimeSeriesApply,
    ProductionPlanTimeSeriesApplyList,
    ProductionPlanTimeSeriesFields,
    ProductionPlanTimeSeriesList,
    ProductionPlanTimeSeriesTextFields,
)
from ._reason import Reason, ReasonApply, ReasonApplyList, ReasonFields, ReasonList, ReasonTextFields
from ._reserve_bid import (
    ReserveBid,
    ReserveBidApply,
    ReserveBidApplyList,
    ReserveBidFields,
    ReserveBidList,
    ReserveBidTextFields,
)
from ._reserve_bid_time_series import (
    ReserveBidTimeSeries,
    ReserveBidTimeSeriesApply,
    ReserveBidTimeSeriesApplyList,
    ReserveBidTimeSeriesFields,
    ReserveBidTimeSeriesList,
    ReserveBidTimeSeriesTextFields,
)
from ._reserve_scenario import (
    ReserveScenario,
    ReserveScenarioApply,
    ReserveScenarioApplyList,
    ReserveScenarioFields,
    ReserveScenarioList,
    ReserveScenarioTextFields,
)
from ._reservoir import (
    Reservoir,
    ReservoirApply,
    ReservoirApplyList,
    ReservoirFields,
    ReservoirList,
    ReservoirTextFields,
)
from ._rkom_bid import RKOMBid, RKOMBidApply, RKOMBidApplyList, RKOMBidFields, RKOMBidList, RKOMBidTextFields
from ._rkom_bid_combination import (
    RKOMBidCombination,
    RKOMBidCombinationApply,
    RKOMBidCombinationApplyList,
    RKOMBidCombinationFields,
    RKOMBidCombinationList,
    RKOMBidCombinationTextFields,
)
from ._rkom_combination_bid import (
    RKOMCombinationBid,
    RKOMCombinationBidApply,
    RKOMCombinationBidApplyList,
    RKOMCombinationBidFields,
    RKOMCombinationBidList,
    RKOMCombinationBidTextFields,
)
from ._rkom_market import (
    RKOMMarket,
    RKOMMarketApply,
    RKOMMarketApplyList,
    RKOMMarketFields,
    RKOMMarketList,
    RKOMMarketTextFields,
)
from ._rkom_process import (
    RKOMProcess,
    RKOMProcessApply,
    RKOMProcessApplyList,
    RKOMProcessFields,
    RKOMProcessList,
    RKOMProcessTextFields,
)
from ._scenario import Scenario, ScenarioApply, ScenarioApplyList, ScenarioFields, ScenarioList, ScenarioTextFields
from ._scenario_mapping import (
    ScenarioMapping,
    ScenarioMappingApply,
    ScenarioMappingApplyList,
    ScenarioMappingFields,
    ScenarioMappingList,
    ScenarioMappingTextFields,
)
from ._scenario_template import (
    ScenarioTemplate,
    ScenarioTemplateApply,
    ScenarioTemplateApplyList,
    ScenarioTemplateFields,
    ScenarioTemplateList,
    ScenarioTemplateTextFields,
)
from ._series import Series, SeriesApply, SeriesApplyList, SeriesFields, SeriesList
from ._shop_transformation import (
    ShopTransformation,
    ShopTransformationApply,
    ShopTransformationApplyList,
    ShopTransformationFields,
    ShopTransformationList,
    ShopTransformationTextFields,
)
from ._time_interval import TimeInterval, TimeIntervalApply, TimeIntervalApplyList, TimeIntervalFields, TimeIntervalList
from ._value_transformation import (
    ValueTransformation,
    ValueTransformationApply,
    ValueTransformationApplyList,
    ValueTransformationFields,
    ValueTransformationList,
    ValueTransformationTextFields,
)
from ._watercourse import (
    Watercourse,
    WatercourseApply,
    WatercourseApplyList,
    WatercourseFields,
    WatercourseList,
    WatercourseTextFields,
)
from ._watercourse_shop import (
    WatercourseShop,
    WatercourseShopApply,
    WatercourseShopApplyList,
    WatercourseShopFields,
    WatercourseShopList,
)

BenchmarkBidApply.model_rebuild()
BenchmarkProcessApply.model_rebuild()
BidDocumentHeaderApply.model_rebuild()
BidTimeSeriesApply.model_rebuild()
DayAheadBidApply.model_rebuild()
DayAheadProcessApply.model_rebuild()
InputTimeSeriesMappingApply.model_rebuild()
OutputContainerApply.model_rebuild()
PeriodsApply.model_rebuild()
PlantApply.model_rebuild()
PriceAreaApply.model_rebuild()
RKOMBidApply.model_rebuild()
RKOMBidCombinationApply.model_rebuild()
RKOMProcessApply.model_rebuild()
ReserveBidApply.model_rebuild()
ReserveBidTimeSeriesApply.model_rebuild()
ScenarioApply.model_rebuild()
ScenarioMappingApply.model_rebuild()
ScenarioTemplateApply.model_rebuild()
SeriesApply.model_rebuild()
ShopTransformationApply.model_rebuild()
WatercourseApply.model_rebuild()

__all__ = [
    "DomainModel",
    "DomainModelApply",
    "BenchmarkBid",
    "BenchmarkBidApply",
    "BenchmarkBidList",
    "BenchmarkBidApplyList",
    "BenchmarkBidFields",
    "BenchmarkBidTextFields",
    "BenchmarkProcess",
    "BenchmarkProcessApply",
    "BenchmarkProcessList",
    "BenchmarkProcessApplyList",
    "BenchmarkProcessFields",
    "BenchmarkProcessTextFields",
    "BidCurves",
    "BidCurvesApply",
    "BidCurvesList",
    "BidCurvesApplyList",
    "BidCurvesFields",
    "BidCurvesTextFields",
    "BidDocumentHeader",
    "BidDocumentHeaderApply",
    "BidDocumentHeaderList",
    "BidDocumentHeaderApplyList",
    "BidDocumentHeaderFields",
    "BidDocumentHeaderTextFields",
    "BidMatrixGenerator",
    "BidMatrixGeneratorApply",
    "BidMatrixGeneratorList",
    "BidMatrixGeneratorApplyList",
    "BidMatrixGeneratorFields",
    "BidMatrixGeneratorTextFields",
    "BidTimeSeries",
    "BidTimeSeriesApply",
    "BidTimeSeriesList",
    "BidTimeSeriesApplyList",
    "BidTimeSeriesFields",
    "BidTimeSeriesTextFields",
    "CommandConfig",
    "CommandConfigApply",
    "CommandConfigList",
    "CommandConfigApplyList",
    "CommandConfigFields",
    "CommandConfigTextFields",
    "DateTimeInterval",
    "DateTimeIntervalApply",
    "DateTimeIntervalList",
    "DateTimeIntervalApplyList",
    "DateTimeIntervalFields",
    "DateTransformation",
    "DateTransformationApply",
    "DateTransformationList",
    "DateTransformationApplyList",
    "DateTransformationFields",
    "DateTransformationTextFields",
    "DayAheadBid",
    "DayAheadBidApply",
    "DayAheadBidList",
    "DayAheadBidApplyList",
    "DayAheadBidFields",
    "DayAheadBidTextFields",
    "DayAheadProcess",
    "DayAheadProcessApply",
    "DayAheadProcessList",
    "DayAheadProcessApplyList",
    "DayAheadProcessFields",
    "DayAheadProcessTextFields",
    "Duration",
    "DurationApply",
    "DurationList",
    "DurationApplyList",
    "DurationFields",
    "DurationTextFields",
    "Generator",
    "GeneratorApply",
    "GeneratorList",
    "GeneratorApplyList",
    "GeneratorFields",
    "GeneratorTextFields",
    "InputTimeSeriesMapping",
    "InputTimeSeriesMappingApply",
    "InputTimeSeriesMappingList",
    "InputTimeSeriesMappingApplyList",
    "InputTimeSeriesMappingFields",
    "InputTimeSeriesMappingTextFields",
    "MBADomain",
    "MBADomainApply",
    "MBADomainList",
    "MBADomainApplyList",
    "MBADomainFields",
    "MBADomainTextFields",
    "MarketAgreement",
    "MarketAgreementApply",
    "MarketAgreementList",
    "MarketAgreementApplyList",
    "MarketAgreementFields",
    "MarketAgreementTextFields",
    "MarketParticipant",
    "MarketParticipantApply",
    "MarketParticipantList",
    "MarketParticipantApplyList",
    "MarketParticipantFields",
    "MarketParticipantTextFields",
    "NordPoolMarket",
    "NordPoolMarketApply",
    "NordPoolMarketList",
    "NordPoolMarketApplyList",
    "NordPoolMarketFields",
    "NordPoolMarketTextFields",
    "OutputContainer",
    "OutputContainerApply",
    "OutputContainerList",
    "OutputContainerApplyList",
    "OutputContainerFields",
    "OutputContainerTextFields",
    "OutputMapping",
    "OutputMappingApply",
    "OutputMappingList",
    "OutputMappingApplyList",
    "OutputMappingFields",
    "OutputMappingTextFields",
    "Periods",
    "PeriodsApply",
    "PeriodsList",
    "PeriodsApplyList",
    "PeriodsFields",
    "PeriodsTextFields",
    "Plant",
    "PlantApply",
    "PlantList",
    "PlantApplyList",
    "PlantFields",
    "PlantTextFields",
    "Point",
    "PointApply",
    "PointList",
    "PointApplyList",
    "PointFields",
    "PriceArea",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
    "ProductionPlanTimeSeries",
    "ProductionPlanTimeSeriesApply",
    "ProductionPlanTimeSeriesList",
    "ProductionPlanTimeSeriesApplyList",
    "ProductionPlanTimeSeriesFields",
    "ProductionPlanTimeSeriesTextFields",
    "RKOMBid",
    "RKOMBidApply",
    "RKOMBidList",
    "RKOMBidApplyList",
    "RKOMBidFields",
    "RKOMBidTextFields",
    "RKOMBidCombination",
    "RKOMBidCombinationApply",
    "RKOMBidCombinationList",
    "RKOMBidCombinationApplyList",
    "RKOMBidCombinationFields",
    "RKOMBidCombinationTextFields",
    "RKOMCombinationBid",
    "RKOMCombinationBidApply",
    "RKOMCombinationBidList",
    "RKOMCombinationBidApplyList",
    "RKOMCombinationBidFields",
    "RKOMCombinationBidTextFields",
    "RKOMMarket",
    "RKOMMarketApply",
    "RKOMMarketList",
    "RKOMMarketApplyList",
    "RKOMMarketFields",
    "RKOMMarketTextFields",
    "RKOMProcess",
    "RKOMProcessApply",
    "RKOMProcessList",
    "RKOMProcessApplyList",
    "RKOMProcessFields",
    "RKOMProcessTextFields",
    "Reason",
    "ReasonApply",
    "ReasonList",
    "ReasonApplyList",
    "ReasonFields",
    "ReasonTextFields",
    "ReserveBid",
    "ReserveBidApply",
    "ReserveBidList",
    "ReserveBidApplyList",
    "ReserveBidFields",
    "ReserveBidTextFields",
    "ReserveBidTimeSeries",
    "ReserveBidTimeSeriesApply",
    "ReserveBidTimeSeriesList",
    "ReserveBidTimeSeriesApplyList",
    "ReserveBidTimeSeriesFields",
    "ReserveBidTimeSeriesTextFields",
    "ReserveScenario",
    "ReserveScenarioApply",
    "ReserveScenarioList",
    "ReserveScenarioApplyList",
    "ReserveScenarioFields",
    "ReserveScenarioTextFields",
    "Reservoir",
    "ReservoirApply",
    "ReservoirList",
    "ReservoirApplyList",
    "ReservoirFields",
    "ReservoirTextFields",
    "Scenario",
    "ScenarioApply",
    "ScenarioList",
    "ScenarioApplyList",
    "ScenarioFields",
    "ScenarioTextFields",
    "ScenarioMapping",
    "ScenarioMappingApply",
    "ScenarioMappingList",
    "ScenarioMappingApplyList",
    "ScenarioMappingFields",
    "ScenarioMappingTextFields",
    "ScenarioTemplate",
    "ScenarioTemplateApply",
    "ScenarioTemplateList",
    "ScenarioTemplateApplyList",
    "ScenarioTemplateFields",
    "ScenarioTemplateTextFields",
    "Series",
    "SeriesApply",
    "SeriesList",
    "SeriesApplyList",
    "SeriesFields",
    "ShopTransformation",
    "ShopTransformationApply",
    "ShopTransformationList",
    "ShopTransformationApplyList",
    "ShopTransformationFields",
    "ShopTransformationTextFields",
    "TimeInterval",
    "TimeIntervalApply",
    "TimeIntervalList",
    "TimeIntervalApplyList",
    "TimeIntervalFields",
    "ValueTransformation",
    "ValueTransformationApply",
    "ValueTransformationList",
    "ValueTransformationApplyList",
    "ValueTransformationFields",
    "ValueTransformationTextFields",
    "Watercourse",
    "WatercourseApply",
    "WatercourseList",
    "WatercourseApplyList",
    "WatercourseFields",
    "WatercourseTextFields",
    "WatercourseShop",
    "WatercourseShopApply",
    "WatercourseShopList",
    "WatercourseShopApplyList",
    "WatercourseShopFields",
]



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = [
    "DateTransformation",
    "DateTransformationApply",
    "DateTransformationList",
    "DateTransformationApplyList",
    "DateTransformationFields",
    "DateTransformationTextFields",
]


DateTransformationTextFields = Literal["transformation", "args"]
DateTransformationFields = Literal["transformation", "args", "kwargs"]

_DATETRANSFORMATION_PROPERTIES_BY_FIELD = {
    "transformation": "transformation",
    "args": "args",
    "kwargs": "kwargs",
}


class DateTransformation(DomainModel):
    space: str = "power-ops"
    transformation: Optional[str] = None
    args: Optional[list[str]] = None
    kwargs: Optional[dict] = None

    def as_apply(self) -> DateTransformationApply:
        return DateTransformationApply(
            external_id=self.external_id,
            transformation=self.transformation,
            args=self.args,
            kwargs=self.kwargs,
        )


class DateTransformationApply(DomainModelApply):
    space: str = "power-ops"
    transformation: Optional[str] = None
    args: Optional[list[str]] = None
    kwargs: Optional[dict] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.transformation is not None:
            properties["transformation"] = self.transformation
        if self.args is not None:
            properties["args"] = self.args
        if self.kwargs is not None:
            properties["kwargs"] = self.kwargs
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "DateTransformation"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class DateTransformationList(TypeList[DateTransformation]):
    _NODE = DateTransformation

    def as_apply(self) -> DateTransformationApplyList:
        return DateTransformationApplyList([node.as_apply() for node in self.data])


class DateTransformationApplyList(TypeApplyList[DateTransformationApply]):
    _NODE = DateTransformationApply



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._bid_document_header import BidDocumentHeaderApply

__all__ = [
    "BidInterval",
    "BidIntervalApply",
    "BidIntervalList",
    "BidIntervalApplyList",
    "BidIntervalFields",
    "BidIntervalTextFields",
]


BidIntervalTextFields = Literal["start", "end"]
BidIntervalFields = Literal["start", "end"]

_BIDINTERVAL_PROPERTIES_BY_FIELD = {
    "start": "Start",
    "end": "End",
}


class BidInterval(DomainModel):
    space: str = "power-ops"
    start: Optional[list[str]] = Field(None, alias="Start")
    end: Optional[list[str]] = Field(None, alias="End")
    parent: Optional[list[str]] = None

    def as_apply(self) -> BidIntervalApply:
        return BidIntervalApply(
            external_id=self.external_id,
            start=self.start,
            end=self.end,
            parent=self.parent,
        )


class BidIntervalApply(DomainModelApply):
    space: str = "power-ops"
    start: Optional[list[str]] = Field(None, alias="Start")
    end: Optional[list[str]] = Field(None, alias="End")
    parent: Union[list[BidDocumentHeaderApply], list[str], None] = Field(default=None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.start is not None:
            properties["Start"] = self.start
        if self.end is not None:
            properties["End"] = self.end
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "BidInterval"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for parent in self.parent or []:
            edge = self._create_parent_edge(parent)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(parent, DomainModelApply):
                instances = parent._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_parent_edge(self, parent: Union[str, BidDocumentHeaderApply]) -> dm.EdgeApply:
        if isinstance(parent, str):
            end_node_ext_id = parent
        elif isinstance(parent, DomainModelApply):
            end_node_ext_id = parent.external_id
        else:
            raise TypeError(f"Expected str or BidDocumentHeaderApply, got {type(parent)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "BidInterval.parent"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class BidIntervalList(TypeList[BidInterval]):
    _NODE = BidInterval

    def as_apply(self) -> BidIntervalApplyList:
        return BidIntervalApplyList([node.as_apply() for node in self.data])


class BidIntervalApplyList(TypeApplyList[BidIntervalApply]):
    _NODE = BidIntervalApply



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = [
    "WatercourseShop",
    "WatercourseShopApply",
    "WatercourseShopList",
    "WatercourseShopApplyList",
    "WatercourseShopFields",
]
WatercourseShopFields = Literal["penalty_limit"]

_WATERCOURSESHOP_PROPERTIES_BY_FIELD = {
    "penalty_limit": "penaltyLimit",
}


class WatercourseShop(DomainModel):
    space: str = "power-ops"
    penalty_limit: Optional[float] = Field(None, alias="penaltyLimit")

    def as_apply(self) -> WatercourseShopApply:
        return WatercourseShopApply(
            external_id=self.external_id,
            penalty_limit=self.penalty_limit,
        )


class WatercourseShopApply(DomainModelApply):
    space: str = "power-ops"
    penalty_limit: Optional[float] = Field(None, alias="penaltyLimit")

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.penalty_limit is not None:
            properties["penaltyLimit"] = self.penalty_limit
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "WatercourseShop"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class WatercourseShopList(TypeList[WatercourseShop]):
    _NODE = WatercourseShop

    def as_apply(self) -> WatercourseShopApplyList:
        return WatercourseShopApplyList([node.as_apply() for node in self.data])


class WatercourseShopApplyList(TypeApplyList[WatercourseShopApply]):
    _NODE = WatercourseShopApply



from __future__ import annotations

from cognite.powerops import PowerOpsClient
from cognite.powerops.resync.models.base import DataModel, T_Model

from . import dms


class CapacityBidModel(DataModel):
    source_model = dms.CapacitySourceModel
    dms_model = dms.CapacityModel

    @classmethod
    def from_cdf(cls: type[T_Model], client: PowerOpsClient, data_set_external_id: str) -> T_Model:
        raise NotImplementedError

    def standardize(self) -> None:
        raise NotImplementedError



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.data_classes import (
    Scenario,
    ScenarioApply,
    ScenarioApplyList,
    ScenarioFields,
    ScenarioList,
    ScenarioTextFields,
)
from cognite.powerops.client._generated.data_classes._scenario import _SCENARIO_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, Aggregations, TypeAPI


class ScenarioAPI(TypeAPI[Scenario, ScenarioApply, ScenarioList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Scenario,
            class_apply_type=ScenarioApply,
            class_list=ScenarioList,
        )
        self._view_id = view_id

    def apply(
        self, scenario: ScenarioApply | Sequence[ScenarioApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(scenario, ScenarioApply):
            instances = scenario.to_instances_apply()
        else:
            instances = ScenarioApplyList(scenario).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Scenario:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ScenarioList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Scenario | ScenarioList:
        if isinstance(external_id, str):
            return self._retrieve((self._sources.space, external_id))
        else:
            return self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

    def search(
        self,
        query: str,
        properties: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ScenarioList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            template,
            mapping,
            commands,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _SCENARIO_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ScenarioFields | Sequence[ScenarioFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ScenarioFields | Sequence[ScenarioFields] | None = None,
        group_by: ScenarioFields | Sequence[ScenarioFields] = None,
        query: str | None = None,
        search_properties: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: ScenarioFields | Sequence[ScenarioFields] | None = None,
        group_by: ScenarioFields | Sequence[ScenarioFields] | None = None,
        query: str | None = None,
        search_property: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            template,
            mapping,
            commands,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _SCENARIO_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ScenarioFields,
        interval: float,
        query: str | None = None,
        search_property: ScenarioTextFields | Sequence[ScenarioTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            template,
            mapping,
            commands,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _SCENARIO_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ScenarioList:
        filter_ = _create_filter(
            self._view_id,
            name,
            name_prefix,
            template,
            mapping,
            commands,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    template: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    mapping: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    commands: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if template and isinstance(template, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("template"), value={"space": "power-ops", "externalId": template})
        )
    if template and isinstance(template, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("template"), value={"space": template[0], "externalId": template[1]}
            )
        )
    if template and isinstance(template, list) and isinstance(template[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("template"),
                values=[{"space": "power-ops", "externalId": item} for item in template],
            )
        )
    if template and isinstance(template, list) and isinstance(template[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("template"),
                values=[{"space": item[0], "externalId": item[1]} for item in template],
            )
        )
    if mapping and isinstance(mapping, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("mapping"), value={"space": "power-ops", "externalId": mapping})
        )
    if mapping and isinstance(mapping, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("mapping"), value={"space": mapping[0], "externalId": mapping[1]})
        )
    if mapping and isinstance(mapping, list) and isinstance(mapping[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("mapping"),
                values=[{"space": "power-ops", "externalId": item} for item in mapping],
            )
        )
    if mapping and isinstance(mapping, list) and isinstance(mapping[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("mapping"),
                values=[{"space": item[0], "externalId": item[1]} for item in mapping],
            )
        )
    if commands and isinstance(commands, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("commands"), value={"space": "power-ops", "externalId": commands})
        )
    if commands and isinstance(commands, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("commands"), value={"space": commands[0], "externalId": commands[1]}
            )
        )
    if commands and isinstance(commands, list) and isinstance(commands[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("commands"),
                values=[{"space": "power-ops", "externalId": item} for item in commands],
            )
        )
    if commands and isinstance(commands, list) and isinstance(commands[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("commands"),
                values=[{"space": item[0], "externalId": item[1]} for item in commands],
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = [
    "MarketParticipant",
    "MarketParticipantApply",
    "MarketParticipantList",
    "MarketParticipantApplyList",
    "MarketParticipantFields",
    "MarketParticipantTextFields",
]


MarketParticipantTextFields = Literal["m_rid", "role"]
MarketParticipantFields = Literal["m_rid", "role"]

_MARKETPARTICIPANT_PROPERTIES_BY_FIELD = {
    "m_rid": "mRID",
    "role": "role",
}


class MarketParticipant(DomainModel):
    space: str = "power-ops"
    m_rid: Optional[str] = Field(None, alias="mRID")
    role: Optional[str] = None

    def as_apply(self) -> MarketParticipantApply:
        return MarketParticipantApply(
            external_id=self.external_id,
            m_rid=self.m_rid,
            role=self.role,
        )


class MarketParticipantApply(DomainModelApply):
    space: str = "power-ops"
    m_rid: Optional[str] = Field(None, alias="mRID")
    role: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.m_rid is not None:
            properties["mRID"] = self.m_rid
        if self.role is not None:
            properties["role"] = self.role
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "MarketParticipant"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class MarketParticipantList(TypeList[MarketParticipant]):
    _NODE = MarketParticipant

    def as_apply(self) -> MarketParticipantApplyList:
        return MarketParticipantApplyList([node.as_apply() for node in self.data])


class MarketParticipantApplyList(TypeApplyList[MarketParticipantApply]):
    _NODE = MarketParticipantApply



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.data_classes import (
    BidMatrixGenerator,
    BidMatrixGeneratorApply,
    BidMatrixGeneratorApplyList,
    BidMatrixGeneratorFields,
    BidMatrixGeneratorList,
    BidMatrixGeneratorTextFields,
)
from cognite.powerops.client._generated.data_classes._bid_matrix_generator import (
    _BIDMATRIXGENERATOR_PROPERTIES_BY_FIELD,
)

from ._core import DEFAULT_LIMIT_READ, Aggregations, TypeAPI


class BidMatrixGeneratorAPI(TypeAPI[BidMatrixGenerator, BidMatrixGeneratorApply, BidMatrixGeneratorList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=BidMatrixGenerator,
            class_apply_type=BidMatrixGeneratorApply,
            class_list=BidMatrixGeneratorList,
        )
        self._view_id = view_id

    def apply(
        self, bid_matrix_generator: BidMatrixGeneratorApply | Sequence[BidMatrixGeneratorApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(bid_matrix_generator, BidMatrixGeneratorApply):
            instances = bid_matrix_generator.to_instances_apply()
        else:
            instances = BidMatrixGeneratorApplyList(bid_matrix_generator).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> BidMatrixGenerator:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> BidMatrixGeneratorList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> BidMatrixGenerator | BidMatrixGeneratorList:
        if isinstance(external_id, str):
            return self._retrieve((self._sources.space, external_id))
        else:
            return self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

    def search(
        self,
        query: str,
        properties: BidMatrixGeneratorTextFields | Sequence[BidMatrixGeneratorTextFields] | None = None,
        shop_plant: str | list[str] | None = None,
        shop_plant_prefix: str | None = None,
        methods: str | list[str] | None = None,
        methods_prefix: str | None = None,
        function_external_id: str | list[str] | None = None,
        function_external_id_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> BidMatrixGeneratorList:
        filter_ = _create_filter(
            self._view_id,
            shop_plant,
            shop_plant_prefix,
            methods,
            methods_prefix,
            function_external_id,
            function_external_id_prefix,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _BIDMATRIXGENERATOR_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidMatrixGeneratorFields | Sequence[BidMatrixGeneratorFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: BidMatrixGeneratorTextFields | Sequence[BidMatrixGeneratorTextFields] | None = None,
        shop_plant: str | list[str] | None = None,
        shop_plant_prefix: str | None = None,
        methods: str | list[str] | None = None,
        methods_prefix: str | None = None,
        function_external_id: str | list[str] | None = None,
        function_external_id_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidMatrixGeneratorFields | Sequence[BidMatrixGeneratorFields] | None = None,
        group_by: BidMatrixGeneratorFields | Sequence[BidMatrixGeneratorFields] = None,
        query: str | None = None,
        search_properties: BidMatrixGeneratorTextFields | Sequence[BidMatrixGeneratorTextFields] | None = None,
        shop_plant: str | list[str] | None = None,
        shop_plant_prefix: str | None = None,
        methods: str | list[str] | None = None,
        methods_prefix: str | None = None,
        function_external_id: str | list[str] | None = None,
        function_external_id_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidMatrixGeneratorFields | Sequence[BidMatrixGeneratorFields] | None = None,
        group_by: BidMatrixGeneratorFields | Sequence[BidMatrixGeneratorFields] | None = None,
        query: str | None = None,
        search_property: BidMatrixGeneratorTextFields | Sequence[BidMatrixGeneratorTextFields] | None = None,
        shop_plant: str | list[str] | None = None,
        shop_plant_prefix: str | None = None,
        methods: str | list[str] | None = None,
        methods_prefix: str | None = None,
        function_external_id: str | list[str] | None = None,
        function_external_id_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            shop_plant,
            shop_plant_prefix,
            methods,
            methods_prefix,
            function_external_id,
            function_external_id_prefix,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _BIDMATRIXGENERATOR_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: BidMatrixGeneratorFields,
        interval: float,
        query: str | None = None,
        search_property: BidMatrixGeneratorTextFields | Sequence[BidMatrixGeneratorTextFields] | None = None,
        shop_plant: str | list[str] | None = None,
        shop_plant_prefix: str | None = None,
        methods: str | list[str] | None = None,
        methods_prefix: str | None = None,
        function_external_id: str | list[str] | None = None,
        function_external_id_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            shop_plant,
            shop_plant_prefix,
            methods,
            methods_prefix,
            function_external_id,
            function_external_id_prefix,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _BIDMATRIXGENERATOR_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        shop_plant: str | list[str] | None = None,
        shop_plant_prefix: str | None = None,
        methods: str | list[str] | None = None,
        methods_prefix: str | None = None,
        function_external_id: str | list[str] | None = None,
        function_external_id_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> BidMatrixGeneratorList:
        filter_ = _create_filter(
            self._view_id,
            shop_plant,
            shop_plant_prefix,
            methods,
            methods_prefix,
            function_external_id,
            function_external_id_prefix,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    shop_plant: str | list[str] | None = None,
    shop_plant_prefix: str | None = None,
    methods: str | list[str] | None = None,
    methods_prefix: str | None = None,
    function_external_id: str | list[str] | None = None,
    function_external_id_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if shop_plant and isinstance(shop_plant, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopPlant"), value=shop_plant))
    if shop_plant and isinstance(shop_plant, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopPlant"), values=shop_plant))
    if shop_plant_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopPlant"), value=shop_plant_prefix))
    if methods and isinstance(methods, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("methods"), value=methods))
    if methods and isinstance(methods, list):
        filters.append(dm.filters.In(view_id.as_property_ref("methods"), values=methods))
    if methods_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("methods"), value=methods_prefix))
    if function_external_id and isinstance(function_external_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionExternalId"), value=function_external_id))
    if function_external_id and isinstance(function_external_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionExternalId"), values=function_external_id))
    if function_external_id_prefix:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("functionExternalId"), value=function_external_id_prefix)
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._bid_curves import BidCurvesApply
    from ._time_interval import TimeIntervalApply

__all__ = ["Periods", "PeriodsApply", "PeriodsList", "PeriodsApplyList", "PeriodsFields", "PeriodsTextFields"]


PeriodsTextFields = Literal["resolution"]
PeriodsFields = Literal["resolution"]

_PERIODS_PROPERTIES_BY_FIELD = {
    "resolution": "Resolution",
}


class Periods(DomainModel):
    space: str = "power-ops"
    resolution: Optional[str] = Field(None, alias="Resolution")
    time_interval: Optional[str] = Field(None, alias="TimeInterval")
    bid_curves: Optional[list[str]] = Field(None, alias="BidCurves")

    def as_apply(self) -> PeriodsApply:
        return PeriodsApply(
            external_id=self.external_id,
            resolution=self.resolution,
            time_interval=self.time_interval,
            bid_curves=self.bid_curves,
        )


class PeriodsApply(DomainModelApply):
    space: str = "power-ops"
    resolution: Optional[str] = Field(None, alias="Resolution")
    time_interval: Union[TimeIntervalApply, str, None] = Field(None, repr=False, alias="TimeInterval")
    bid_curves: Union[list[BidCurvesApply], list[str], None] = Field(default=None, repr=False, alias="BidCurves")

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.resolution is not None:
            properties["Resolution"] = self.resolution
        if self.time_interval is not None:
            properties["TimeInterval"] = {
                "space": "power-ops",
                "externalId": self.time_interval
                if isinstance(self.time_interval, str)
                else self.time_interval.external_id,
            }
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "Periods"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for bid_curve in self.bid_curves or []:
            edge = self._create_bid_curve_edge(bid_curve)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(bid_curve, DomainModelApply):
                instances = bid_curve._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.time_interval, DomainModelApply):
            instances = self.time_interval._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_bid_curve_edge(self, bid_curve: Union[str, BidCurvesApply]) -> dm.EdgeApply:
        if isinstance(bid_curve, str):
            end_node_ext_id = bid_curve
        elif isinstance(bid_curve, DomainModelApply):
            end_node_ext_id = bid_curve.external_id
        else:
            raise TypeError(f"Expected str or BidCurvesApply, got {type(bid_curve)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "ReserveBidTimeSeries.BidCurves"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class PeriodsList(TypeList[Periods]):
    _NODE = Periods

    def as_apply(self) -> PeriodsApplyList:
        return PeriodsApplyList([node.as_apply() for node in self.data])


class PeriodsApplyList(TypeApplyList[PeriodsApply]):
    _NODE = PeriodsApply



from __future__ import annotations

import datetime
from typing import Literal, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = [
    "MarketAgreement",
    "MarketAgreementApply",
    "MarketAgreementList",
    "MarketAgreementApplyList",
    "MarketAgreementFields",
    "MarketAgreementTextFields",
]


MarketAgreementTextFields = Literal["m_rid", "type"]
MarketAgreementFields = Literal["m_rid", "type", "created_timestamp"]

_MARKETAGREEMENT_PROPERTIES_BY_FIELD = {
    "m_rid": "mRID",
    "type": "type",
    "created_timestamp": "createdTimestamp",
}


class MarketAgreement(DomainModel):
    space: str = "power-ops"
    m_rid: Optional[str] = Field(None, alias="mRID")
    type: Optional[str] = None
    created_timestamp: Optional[datetime.datetime] = Field(None, alias="createdTimestamp")

    def as_apply(self) -> MarketAgreementApply:
        return MarketAgreementApply(
            external_id=self.external_id,
            m_rid=self.m_rid,
            type=self.type,
            created_timestamp=self.created_timestamp,
        )


class MarketAgreementApply(DomainModelApply):
    space: str = "power-ops"
    m_rid: Optional[str] = Field(None, alias="mRID")
    type: Optional[str] = None
    created_timestamp: Optional[datetime.datetime] = Field(None, alias="createdTimestamp")

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.m_rid is not None:
            properties["mRID"] = self.m_rid
        if self.type is not None:
            properties["type"] = self.type
        if self.created_timestamp is not None:
            properties["createdTimestamp"] = self.created_timestamp.isoformat(timespec="milliseconds")
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "MarketAgreement"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class MarketAgreementList(TypeList[MarketAgreement]):
    _NODE = MarketAgreement

    def as_apply(self) -> MarketAgreementApplyList:
        return MarketAgreementApplyList([node.as_apply() for node in self.data])


class MarketAgreementApplyList(TypeApplyList[MarketAgreementApply]):
    _NODE = MarketAgreementApply



from __future__ import annotations

from typing import ClassVar

from cognite.client.data_classes.data_modeling import ContainerId
from pydantic import Field, field_validator

from cognite.powerops.client._generated.data_classes._core import DomainModelApply
from cognite.powerops.client.data_classes import (
    GeneratorApply,
    PlantApply,
    PriceAreaApply,
    ReservoirApply,
    WatercourseApply,
    WatercourseShopApply,
)
from cognite.powerops.client.powerops_client import PowerOpsClient
from cognite.powerops.resync.models.base import CDFSequence, DataModel, PowerOpsGraphQLModel, T_Model

from .graphql_schemas import GRAPHQL_MODELS


class ProductionModelDM(DataModel):
    graph_ql: ClassVar[PowerOpsGraphQLModel] = GRAPHQL_MODELS["production"]
    cls_by_container: ClassVar[dict[ContainerId, type[DomainModelApply]]] = {
        ContainerId("power-ops", "PriceArea"): PriceAreaApply,
        ContainerId("power-ops", "Watercourse"): WatercourseApply,
        ContainerId("power-ops", "Plant"): PlantApply,
        ContainerId("power-ops", "Generator"): GeneratorApply,
        ContainerId("power-ops", "Reservoir"): ReservoirApply,
        ContainerId("power-ops", "WatercourseShop"): WatercourseShopApply,
    }
    cdf_sequences: list[CDFSequence] = Field(default_factory=list)
    price_areas: list[PriceAreaApply] = Field(default_factory=list)
    watercourses: list[WatercourseApply] = Field(default_factory=list)
    plants: list[PlantApply] = Field(default_factory=list)
    generators: list[GeneratorApply] = Field(default_factory=list)
    reservoirs: list[ReservoirApply] = Field(default_factory=list)

    @field_validator("cdf_sequences", "price_areas", "watercourses", "plants", "generators", "reservoirs", mode="after")
    def ordering_list(cls, value: list) -> list:
        return sorted(value, key=lambda x: x.external_id)

    @classmethod
    def from_cdf(cls: type[T_Model], client: PowerOpsClient, data_set_external_id: str) -> T_Model:
        production = client.production
        price_areas = production.price_area.list(limit=-1)
        watercourses = production.watercourse.list(limit=-1)
        plants = production.plant.list(limit=-1)
        generators = production.generator.list(limit=-1)
        reservoirs = production.reservoir.list(limit=-1)
        watercourse_shop = production.watercourse_shop.list(limit=-1)
        sequence_external_ids = [
            external_id
            for generator in generators
            for external_id in [generator.generator_efficiency_curve, generator.turbine_efficiency_curve]
        ]
        if sequence_external_ids:
            sequences = client.cdf.sequences.retrieve_multiple(
                external_ids=sequence_external_ids, ignore_unknown_ids=True
            )
        else:
            sequences = []

        watercourse_shop_by_external_id = {w.external_id: w.as_apply() for w in watercourse_shop}
        for watercourse in watercourses:
            watercourse.shop = watercourse_shop_by_external_id[watercourse.shop]

        return cls(
            price_areas=list(price_areas.as_apply()),
            watercourses=list(watercourses.as_apply()),
            plants=list(plants.as_apply()),
            generators=list(generators.as_apply()),
            reservoirs=list(reservoirs.as_apply()),
            cdf_sequences=[CDFSequence(sequence=sequence) for sequence in sequences],
        )

    def standardize(self) -> None:
        self.cdf_sequences = self.ordering_list(self.cdf_sequences)
        self.price_areas = self.ordering_list(self.price_areas)
        self.watercourses = self.ordering_list(self.watercourses)
        self.plants = self.ordering_list(self.plants)
        self.generators = self.ordering_list(self.generators)
        self.reservoirs = self.ordering_list(self.reservoirs)
